library(xtable)

log.filename <- "log/classification.log"
latex.filename <- "out/classification/classification.tex"
simulations <- 100000
#simulations <- 5000

source("start.R")
source("include-classification-model-specification.R")
source("include-classification.R")
source("include-marginal-logic.R")

sink(latex.filename)
cat("\\documentclass[a4paper,10pt,oneside,english]{article}\\usepackage{float}\\usepackage{lscape}\\usepackage[pdfstartview=FitH,pdftitle={},pdfauthor={Mikkel Meyer Andersen},pdfsubject={},bookmarksnumbered,pagebackref,bookmarksopen,hyperfootnotes=false]{hyperref}\\urlstyle{same}\\hypersetup{pdfborder=0 0 1}\\input{../../../report/datasetnames}\\begin{document}\\section*{Simulation is based on", simulations, "simulated haplotypes}")
sink()

sink(log.filename)

cat("Number of haplotypes sampled to calculate marginals = ", simulations, "\n\n")

unobserved.prob.matrix <- matrix(NA, 
  nrow=length(dataset.names), 
  ncol=length(classification.model.names[classification.model.use]) + 2, 
  dimnames=list(paste("¤DATASET:", dataset.names, "¤", sep=""), c("Estimate", "CI", classification.model.names[classification.model.use])))

deviance.relative.predicted.matrix <- matrix(NA, 
  nrow=length(dataset.names), 
  ncol=length(classification.model.names[classification.model.use]), 
  dimnames=list(paste("¤DATASET:", dataset.names, "¤", sep=""), classification.model.names[classification.model.use]))

multinomial.probabilities.matrix <- matrix(NA, 
  nrow=length(dataset.names), 
  ncol=length(classification.model.names[classification.model.use]), 
  dimnames=list(paste("¤DATASET:", dataset.names, "¤", sep=""), classification.model.names[classification.model.use]))

multinomial.log.probabilities.matrix <- matrix(NA, 
  nrow=length(dataset.names), 
  ncol=length(classification.model.names[classification.model.use]) + 1, 
  dimnames=list(paste("¤DATASET:", dataset.names, "¤", sep=""), c("Reciprocal", classification.model.names[classification.model.use])))
  
unobserved.prob.estimates <- c("0.364", "0.602", "0.277")
unobserved.prob.intervals <- c("[0.321; 0.408]", "[0.510; 0.694]", "[0.212; 0.342]")


#############################################
####### MARGINAL FUNCTIONS ##################
#############################################
# MARGINALS FUNCTIONS ->
create.marginals.deviance.matrix <- function() {
  return(matrix(NA, 
    nrow=length(dataset.names), 
    ncol=length(classification.model.names[classification.model.use]), 
    dimnames=list(paste("¤DATASET:", dataset.names, "¤", sep=""), classification.model.names[classification.model.use])))
}

create.single.marginals.rejected.loci.matrix <- function() {
  return(matrix(NA, 
    nrow=length(dataset.names), 
    ncol=length(classification.model.names[classification.model.use])+1, 
    dimnames=list(paste("¤DATASET:", dataset.names, "¤", sep=""), c("Maximum", classification.model.names[classification.model.use]))))
}

create.pairwise.marginals.rejected.loci.matrix <- function() {
  return(matrix(NA, 
    nrow=length(dataset.names), 
    ncol=length(classification.model.names[classification.model.use])+1, 
    dimnames=list(paste("¤DATASET:", dataset.names, "¤", sep=""), c("Maximum", classification.model.names[classification.model.use]))))
}
# <- MARGINALS FUNCTIONS

#############################################
####### SINGLE MARGINALS ####################
#############################################
# SINGLE MARGINALS: OBSERVED VS. SIMULATED ->
single.marginals.deviance.sum.obs.vs.sim.matrix <- create.marginals.deviance.matrix()
single.marginals.loci.rejected.obs.vs.sim.matrix <- create.single.marginals.rejected.loci.matrix()
# <- SINGLE MARGINALS: OBSERVED VS. SIMULATED
# SINGLE MARGINALS: SIMULATED VS. PREDICTED ->
single.marginals.deviance.sum.sim.vs.pred.matrix <- create.marginals.deviance.matrix()
single.marginals.loci.rejected.sim.vs.pred.matrix <- create.single.marginals.rejected.loci.matrix()
# <- SINGLE MARGINALS: SIMULATED VS. PREDICTED

#############################################
####### PAIRWISE MARGINALS ##################
#############################################
# PAIRWISE MARGINALS: OBSERVED VS. SIMULATED ->
pairwise.marginals.deviance.sum.obs.vs.sim.matrix <- create.marginals.deviance.matrix()
pairwise.marginals.loci.rejected.obs.vs.sim.matrix <- create.pairwise.marginals.rejected.loci.matrix()
# <- PAIRWISE MARGINALS: OBSERVED VS. SIMULATED
# PAIRWISE MARGINALS: SIMULATED VS. PREDICTED ->
pairwise.marginals.deviance.sum.sim.vs.pred.matrix <- create.marginals.deviance.matrix()
pairwise.marginals.loci.rejected.sim.vs.pred.matrix <- create.pairwise.marginals.rejected.loci.matrix()
# <- PAIRWISE MARGINALS: SIMULATED VS. PREDICTED


jobs.to.do <- length(dataset.names) * length(classification.model.use)
jobs.done <- 0

for (i in 1:length(dataset.names)) {
  dataset.name <- dataset.names[i]  

  cat("Starting classification for dataset \"", dataset.name, "\".\n", sep="")

  # Global assignment: hack, but easier...
  loci.count <- get.loci.count(dataset.name)
  loci <- 1:loci.count
  loci.names <- get.loci.names(dataset.name)

  dataset.compact <- get.compact.dataset(dataset.name)
  dataset.extended <- get.extended.dataset(dataset.name)
  observation.count <- nrow(dataset.extended)

  cat("Number of observations = ", observation.count, "\n", sep="")

  unobserved.prob.matrix[i, 1] <- unobserved.prob.estimates[i]
  unobserved.prob.matrix[i, 2] <- unobserved.prob.intervals[i]

  # The support of the multinomial distribution is the set
  # { (n1, n2, ..., nk) : n1 + n2 + ... + nk = N }
  # which have
  # choose(N + k - 1, k - 1)
  # elements
  # Then if each element has the same probability, each element has
  # 1 / choose(N + k - 1, k - 1)
  # such that
  # log(1 / choose(N + k - 1, k - 1)) 
  # = log(1) - log(choose(N + k - 1, k - 1))
  # = - lchoose(N + k - 1, k - 1)
  loci.levels <- get.loci.levels(dataset.name)
  size.of.H <- prod(sapply(loci.levels, length))
  multinomial.log.probabilities.matrix[i, 1] <- - lchoose(observation.count + size.of.H - 1, size.of.H - 1)

  # MARGINALS FUNCTIONS ->
  create.single.marginals.each.method.matrix <- function () {
    return(matrix(NA, 
      nrow=loci.count, 
      ncol=length(classification.model.names[classification.model.use]), 
      dimnames=list(paste("¤LOCUS:", loci.names, "¤", sep=""), classification.model.names[classification.model.use]))
    )
  }
  # <- MARGINALS FUNCTIONS
  
  # SINGLE MARGINALS: OBSERVED VS. SIMULATED ->
  single.marginals.loci.rejected.obs.vs.sim.matrix[i, 1] <- loci.count - 1
  deviances.each.method.single.marginals.obs.vs.sim <- create.single.marginals.each.method.matrix()
  p.values.each.method.single.marginals.obs.vs.sim <- create.single.marginals.each.method.matrix()
  # <- SINGLE MARGINALS: OBSERVED VS. SIMULATED
  # SINGLE MARGINALS: SIMULATED VS. PREDICTED ->
  single.marginals.loci.rejected.sim.vs.pred.matrix[i, 1] <- loci.count
  deviances.each.method.single.marginals.sim.vs.pred <- create.single.marginals.each.method.matrix()
  p.values.each.method.single.marginals.sim.vs.pred <- create.single.marginals.each.method.matrix()
  # <- SINGLE MARGINALS: SIMULATED VS. PREDICTED

  # PAIRWISE MARGINALS: OBSERVED VS. SIMULATED ->
  pairwise.marginals.loci.rejected.obs.vs.sim.matrix[i, 1] <- choose(loci.count, 2)
  # <- PAIRWISE MARGINALS: OBSERVED VS. SIMULATED
  # PAIRWISE MARGINALS: SIMULATED VS. PREDICTED ->
  pairwise.marginals.loci.rejected.sim.vs.pred.matrix[i, 1] <- choose(loci.count, 2)
  # <- PAIRWISE MARGINALS: SIMULATED VS. PREDICTED

  for (j in 1:length(classification.model.use)) {
    m <- classification.model.use[j]
    classification.model.name <- classification.model.names[m]
    classification.model.function <- classification.model.functions[[m]]
    classification.model.score <- classification.model.scores[[m]]
    classification.model.prediction <- classification.model.predictions[[m]]

    cat("Using classification method \"", classification.model.name, "\".\n", sep="")

    model.fits <- NULL
    
    resultTry <- try(
      model.fits <- get.best.model.fits(dataset.name,
        classification.model.name,
        classification.model.function, 
        classification.model.score)
    )

    if (is.null(model.fits) || inherits(resultTry, "try-error")) {
      unobserved.prob.matrix[i, j+2] <- NA
      cat("Method \"", classification.model.name, "\" failed.\n\n", sep="")
      next
    } 

    locus.not.used <- setdiff(loci, model.fits$loci.used)

    resultProbabilities <- predict.probabilities(dataset.name, model.fits,
      classification.model.name,
      classification.model.function, 
      classification.model.prediction)

    unobserved.prob.matrix[i, j+2] <- round(resultProbabilities$uncovered.prob.mass, 3)
    deviance.relative.predicted.matrix[i, j] <- deviance.relfreq.predicted(dataset.compact$N, resultProbabilities$probs)

    multinomial.probabilities.result <- log.multinomial.probability.with.predicted(dataset.compact$N, resultProbabilities$probs)
    multinomial.probabilities.matrix[i, j] <- multinomial.probabilities.result$p
    multinomial.log.probabilities.matrix[i, j+1] <- multinomial.probabilities.result$log.p

    dataset.simulated <- simulate.haplotypes(dataset.name, model.fits,
      simulations = simulations,
      classification.model.name,
      classification.model.prediction)

    #############################################
    ####### SINGLE MARGINALS ####################
    #############################################
    # SINGLE MARGINALS: OBSERVED VS. SIMULATED ->
    cat("¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤\n")
    cat("OBSERVED VS. SIMULATED:\n")
    cat("¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤\n")
    single.marginals.obs.vs.sim.deviance.analyze <- analyze.single.marginals(dataset.extended, 
      dataset.simulated, "Observed", "Simulated", locus.to.ignore = locus.not.used)

    single.marginals.deviance.sum.obs.vs.sim.matrix[i, j] <- single.marginals.obs.vs.sim.deviance.analyze$deviance.sum
    single.marginals.loci.rejected.obs.vs.sim.matrix[i, j+1] <- single.marginals.obs.vs.sim.deviance.analyze$loci.rejected

    deviances.each.method.single.marginals.obs.vs.sim[loci[-locus.not.used], j] <- single.marginals.obs.vs.sim.deviance.analyze$deviances
    p.values.each.method.single.marginals.obs.vs.sim[loci[-locus.not.used], j] <- single.marginals.obs.vs.sim.deviance.analyze$p.values
    # <- SINGLE MARGINALS: OBSERVED VS. SIMULATED

    # SINGLE MARGINALS: SIMULATED VS. PREDICTED ->
    cat("¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤\n")
    cat("SIMULATED VS. PREDICTED:\n")
    cat("¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤\n")
    # sum(resultProbabilities$probs) = 1 - resultProbabilities$uncovered.prob.mass
    probs.normalized <- resultProbabilities$probs / ( 1 - resultProbabilities$uncovered.prob.mass )
    compact.marginals.model <- dataset.compact

    # Substitute observed N with predicted
    compact.marginals.model$N <- probs.normalized * observation.count
    compact.marginals.model$Freq <- NULL
    single.marginals.sim.vs.pred.deviance.analyze <- analyze.single.marginals(dataset.simulated, 
      compact.marginals.model, "Simulated", "Predicted", locus.to.ignore = NULL)
    #  compact.marginals.model, "Simulated", "Predicted", locus.not.used)

    single.marginals.deviance.sum.sim.vs.pred.matrix[i, j] <- single.marginals.sim.vs.pred.deviance.analyze$deviance.sum
    single.marginals.loci.rejected.sim.vs.pred.matrix[i, j+1] <- single.marginals.sim.vs.pred.deviance.analyze$loci.rejected

    deviances.each.method.single.marginals.sim.vs.pred[loci, j] <- single.marginals.sim.vs.pred.deviance.analyze$deviances
    p.values.each.method.single.marginals.sim.vs.pred[loci, j] <- single.marginals.sim.vs.pred.deviance.analyze$p.values

    #deviances.each.method.single.marginals.sim.vs.pred[loci[-locus.not.used], j] <- single.marginals.sim.vs.pred.deviance.analyze$deviances
    #p.values.each.method.single.marginals.sim.vs.pred[loci[-locus.not.used], j] <- single.marginals.sim.vs.pred.deviance.analyze$p.values
    # <- SINGLE MARGINALS: SIMULATED VS. PREDICTED 


    #############################################
    ####### PAIRWISE MARGINALS ##################
    #############################################    
    # PAIRWISE MARGINALS: OBSERVED VS. SIMULATED ->
    cat("¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤\n")
    cat("OBSERVED VS. SIMULATED:\n")
    cat("¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤\n")
    pairwise.marginals.obs.vs.sim.deviance.analyze <- analyze.pairwise.marginals(dataset.extended, 
      dataset.simulated, "Observed", "Simulated", locus.to.ignore = NULL)

    pairwise.marginals.deviance.sum.obs.vs.sim.matrix[i, j] <- pairwise.marginals.obs.vs.sim.deviance.analyze$deviance.sum
    pairwise.marginals.loci.rejected.obs.vs.sim.matrix[i, j+1] <- pairwise.marginals.obs.vs.sim.deviance.analyze$loci.rejected
    # <- PAIRWISE MARGINALS: OBSERVED VS. SIMULATED

    # PAIRWISE MARGINALS: SIMULATED VS. PREDICTED ->
    cat("¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤\n")
    cat("SIMULATED VS. PREDICTED:\n")
    cat("¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤\n")
    # sum(resultProbabilities$probs) = 1 - resultProbabilities$uncovered.prob.mass
    probs.normalized <- resultProbabilities$probs / ( 1 - resultProbabilities$uncovered.prob.mass )
    compact.marginals.model <- dataset.compact

    # Substitute observed N with predicted
    compact.marginals.model$N <- probs.normalized * observation.count
    compact.marginals.model$Freq <- NULL
    pairwise.marginals.sim.vs.pred.deviance.analyze <- analyze.pairwise.marginals(dataset.simulated, 
      compact.marginals.model, "Simulated", "Predicted", locus.to.ignore = NULL)

    pairwise.marginals.deviance.sum.sim.vs.pred.matrix[i, j] <- pairwise.marginals.sim.vs.pred.deviance.analyze$deviance.sum
    pairwise.marginals.loci.rejected.sim.vs.pred.matrix[i, j+1] <- pairwise.marginals.sim.vs.pred.deviance.analyze$loci.rejected
    # <- PAIRWISE MARGINALS: SIMULATED VS. PREDICTED 
  
    cat("Method \"", classification.model.name, "\" done.\n\n", sep="")

    jobs.done <- jobs.done + 1
    sink()
    cat(round(100 * jobs.done / jobs.to.do, 3), "%\n")
    sink(log.filename, append = TRUE) # start log again
  }

  #############################################
  ####### SINGLE MARGINALS ####################
  #############################################
  # SINGLE MARGINALS: OBSERVED VS. SIMULATED ->
  create.perdataset.output.single.marginals("classification", deviances.each.method.single.marginals.obs.vs.sim, 
    dataset.name, "Deviances", "observed", "simulated")
  create.perdataset.output.single.marginals("classification", p.values.each.method.single.marginals.obs.vs.sim, 
    dataset.name, "P-values", "observed", "simulated")
  # <- SINGLE MARGINALS: OBSERVED VS. SIMULATED
  # SINGLE MARGINALS: SIMULATED VS. PREDICTED ->
  create.perdataset.output.single.marginals("classification", deviances.each.method.single.marginals.sim.vs.pred, 
    dataset.name, "Deviances", "simulated", "predicted")
  create.perdataset.output.single.marginals("classification", p.values.each.method.single.marginals.sim.vs.pred, 
    dataset.name, "P-values", "simulated", "predicted")
  # <- SINGLE MARGINALS: SIMULATED VS. PREDICTED 
  
  cat("Dataset \"", dataset.name, "\" done.\n\n", sep="")
}

cat("Predicted unobserved probability mass:\n")
print(unobserved.prob.matrix)
cat("\n")

unobserved.prob.matrix <- t(unobserved.prob.matrix)
tableOutput <- xtable(unobserved.prob.matrix,
      align=c("l", rep("r", ncol(unobserved.prob.matrix))),
      label=paste("tab:classification-unobserved-probabilities", sep=""), 
      caption=paste("Unobserved probabilities using classification methods. The missing values indicate that the method did not succeed for some reason."))
filename <- paste("out/classification/unobserved-probabilities.tex", sep="")
print(tableOutput, file=filename)
do.replacements(filename)

deviance.relative.predicted.matrix <- t(deviance.relative.predicted.matrix)
tableOutput <- xtable(deviance.relative.predicted.matrix,
      align=c("l", rep("r", ncol(deviance.relative.predicted.matrix))),
      label=paste("tab:classification-deviance-relative-predicted", sep=""), 
      caption=paste("Comparing deviance between relative frequencies and predicted probabilities found using classification methods."))
filename <- paste("out/classification/deviance-relative-predicted.tex", sep="")
print(tableOutput, file=filename)
do.replacements(filename)

multinomial.probabilities.matrix <- t(multinomial.probabilities.matrix)
tableOutput <- xtable(multinomial.probabilities.matrix,
      align=c("l", rep("r", ncol(multinomial.probabilities.matrix))),
      display=c("s", rep("e", ncol(multinomial.probabilities.matrix))),
      label=paste("tab:classification-multinomial-probabilities", sep=""), 
      caption=paste("The probability for the observed counts assuming a multinomial distribution with the predicted probabilities using classification methods as the probabilities parameter."))
filename <- paste("out/classification/multinomial-probabilities.tex", sep="")
print(tableOutput, file=filename)
do.replacements(filename)

multinomial.log.probabilities.matrix <- t(multinomial.log.probabilities.matrix)
tableOutput <- xtable(multinomial.log.probabilities.matrix,
      align=c("l", rep("r", ncol(multinomial.log.probabilities.matrix))),
      display=c("s", rep("e", ncol(multinomial.log.probabilities.matrix))),
      label=paste("tab:classification-multinomial-log-probabilities", sep=""), 
      caption=paste("The log probability for the observed counts assuming a multinomial distribution with the predicted probabilities using classification methods as the probabilities parameter. The reciprocal row is the probability of sampling a certain element in the support for the multinomial distribution assuming uniform sampling probabilities."))
filename <- paste("out/classification/multinomial-log-probabilities.tex", sep="")
print(tableOutput, file=filename)
do.replacements(filename)

#############################################
####### SINGLE MARGINALS ##################
#############################################
# SINGLE MARGINALS: OBSERVED VS. SIMULATED ->
create.overall.output.marginals("classification", "single", single.marginals.deviance.sum.obs.vs.sim.matrix, 
  "deviances-sum", "Sum of deviances", "observed", "simulated", columns.is.int = FALSE)
create.overall.output.marginals("classification", "single", single.marginals.loci.rejected.obs.vs.sim.matrix, 
  "loci-rejected", "Number of loci rejected", "observed", "simulated", columns.is.int = TRUE)
# <- SINGLE MARGINALS: OBSERVED VS. SIMULATED
# SINGLE MARGINALS: SIMULATED VS. PREDICTED ->
create.overall.output.marginals("classification", "single", single.marginals.deviance.sum.sim.vs.pred.matrix, 
  "deviances-sum", "Sum of deviances", "simulated", "predicted", columns.is.int = FALSE)
create.overall.output.marginals("classification", "single", single.marginals.loci.rejected.sim.vs.pred.matrix, 
  "loci-rejected", "Number of loci rejected", "simulated", "predicted", columns.is.int = TRUE)
# <- SINGLE MARGINALS: SIMULATED VS. PREDICTED

#############################################
####### PAIRWISE MARGINALS ##################
#############################################
# PAIRWISE MARGINALS: OBSERVED VS. SIMULATED ->
create.overall.output.marginals("classification", "pairwise", pairwise.marginals.deviance.sum.obs.vs.sim.matrix, 
  "deviances-sum", "Sum of deviances", "observed", "simulated", columns.is.int = FALSE)
create.overall.output.marginals("classification", "pairwise", pairwise.marginals.loci.rejected.obs.vs.sim.matrix, 
  "loci-rejected", "Number of loci rejected", "observed", "simulated", columns.is.int = TRUE)
# <- PAIRWISE MARGINALS: OBSERVED VS. SIMULATED
# PAIRWISE MARGINALS: SIMULATED VS. PREDICTED ->
create.overall.output.marginals("classification", "pairwise", pairwise.marginals.deviance.sum.sim.vs.pred.matrix, 
  "deviances-sum", "Sum of deviances", "simulated", "predicted", columns.is.int = FALSE)
create.overall.output.marginals("classification", "pairwise", pairwise.marginals.loci.rejected.sim.vs.pred.matrix, 
  "loci-rejected", "Number of loci rejected", "simulated", "predicted", columns.is.int = TRUE)
# <- PAIRWISE MARGINALS: SIMULATED VS. PREDICTED
cat("Done.\n")

sink() # sink(log.filename)

sink(latex.filename, append = TRUE)
cat("\\input{unobserved-probabilities}\n\\clearpage\n")
cat("\\input{deviance-relative-predicted}\n\\clearpage\n")
cat("\\input{multinomial-probabilities}\n\\clearpage\n")
cat("\\input{multinomial-log-probabilities}\n")
cat("\\end{document}")
sink()

system("/home/mikl/mat6/R/out/classification/correct-methods-names.sh")
