# If classification.model.number is NULL, then use the model.fits from last time

evidence.two.contributors <- function(dataset.name, classification.model.number, verbose = FALSE, keep.contributors.from.last.run = FALSE) {
  # Global assignment: hack, but easier...
  loci.count <<- get.loci.count(dataset.name)
  loci <<- 1:loci.count
  loci.names <<- get.loci.names(dataset.name)

  dataset.extended <- get.extended.dataset(dataset.name)

  use.last.model <- ifelse(is.null(classification.model.number), TRUE, FALSE)

  if (!use.last.model) {
    classification.model.name <<- classification.model.names[classification.model.number]
    classification.model.function <<- classification.model.functions[[classification.model.number]]
    classification.model.score <<- classification.model.scores[[classification.model.number]]
    classification.model.prediction <<- classification.model.predictions[[classification.model.number]]
  }

  if (is.null(classification.model.name)) stop("is.null(classification.model.name)")
  if (is.null(classification.model.function)) stop("is.null(classification.model.function)")
  if (is.null(classification.model.score)) stop("is.null(classification.model.score)")
  if (is.null(classification.model.prediction)) stop("is.null(classification.model.prediction)")

  if (verbose) {
    cat("Using classification method \"", classification.model.name, "\" on dataset \"", dataset.name, "\" for evidence testing.\n", sep="")
  }

  if (!exists("model.fits") || is.null(model.fits) || !use.last.model) {
    sink('/dev/null') 

    # No need to use try as in classification.R - rpart succeeds on all the datasets
    # Make global if to be reused
    model.fits <<- get.best.model.fits(dataset.name,
      classification.model.name,
      classification.model.function, 
      classification.model.score)
    sink() 
  }

  if (!keep.contributors.from.last.run || !exists("contributor.haplotypes") || is.null(contributor.haplotypes)) {
    contributor.haplotypes <- simulate.haplotypes(dataset.name, model.fits,
      simulations = 2,
      classification.model.name,
      classification.model.prediction)
  }

  haplotype.prob <- function(h) {
    return(predict.haplotypes.probabilities(h, 
      dataset.extended, model.fits,
      classification.model.name,
      classification.model.function, 
      classification.model.prediction))
  }

  is.contributors <- function(h1, h2) {
    if (all(contributor.haplotypes[1, ] == h1) && all(contributor.haplotypes[2, ] == h2))
      return(TRUE)
    if (all(contributor.haplotypes[1, ] == h2) && all(contributor.haplotypes[2, ] == h1))
      return(TRUE)

    return(FALSE)
  }

  contributors.probs <- haplotype.prob(contributor.haplotypes)

  trace <- apply(contributor.haplotypes, 2, unique)
  trace.string <- paste("( ", paste("{", lapply(trace, paste, collapse=", "), "}", collapse=", ", sep=""), ")", sep="")

  if (verbose) {
    cat("Contributors:\n")
    print(contributor.haplotypes)

    cat("Contributors probabilities:\n")
    print(contributors.probs)
    cat("multiplied = ", prod(contributors.probs), "\n")

    cat("Trace T =", trace.string, "\n")
  }

  js <- which(as.numeric(sapply(trace, length)) == 2)
  r <- length(trace)
  k <- r - length(js)

  if (k == r) {
    #stop("Trivial case: no loci in trace with two alleles!")
    return(list(
      contributors.probs = contributors.probs,
      haplotype.universe.size = 1,
      prob.sum = sum(contributors.probs),
      contributors.place = 1 
      ))
  }

  j <- js[1]
  trace.modified <- trace
  trace.modified[[j]] <- trace[[j]][1]
  haplotype.universe <- expand.grid(trace.modified)

  if (verbose) {  
    #cat("Haplotype universe:\n")
    #print(haplotype.universe)

    cat("\n")
    cat("Number of haplotypes to estimate probability for =", 2*2^(r-k-1), "\n")
  }
  
  h1s <- data.frame()
  h2s <- data.frame()

  for (i in 1:nrow(haplotype.universe)) {
    h1 <- as.numeric(haplotype.universe[i, ])   
    h2 <- lapply(1:r, function(i) {
              trace.i <- trace[[i]]
                if (length(trace.i) == 2) trace.i[trace.i != h1[i]] 
                else trace.i
              }
          )

    #cat("h1 = (", paste(h1, collapse=", "), ")\n")
    #cat("h2 = (", paste(h2, collapse=", "), ")\n")

    h1s <- rbind(h1s, h1)
    h2s <- rbind(h2s, h2)
  }

  colnames(h1s) <- colnames(h2s) <- colnames(contributor.haplotypes)

  h1s.prob <- haplotype.prob(h1s)
  h2s.prob <- haplotype.prob(h2s)

  prob.prod <<- h1s.prob * h2s.prob
  prob.prod.sum <- sum(prob.prod)

  arg.max.ordering <- order(prob.prod, decreasing = TRUE)

  contributors.place <- 0

  for (place in 1:nrow(haplotype.universe)) {
    i <- arg.max.ordering[place]
    
    h1 <- as.numeric(h1s[i, ])
    h2 <- as.numeric(h2s[i, ])

    if (is.contributors(h1, h2)) {
      contributors.place <- place
      break
    }
  }

  top.limit <- 5

  if (verbose) {
    cat("\n")
    cat("Top-", top.limit, ":\n", sep="")
    cat("===============================\n")
    
    for (place in 1:top.limit) {
      if (place > nrow(haplotype.universe)) {
        cat("No more combinations...\n")
        break
      }

      i <- arg.max.ordering[place]
      h1 <- as.numeric(h1s[i, ])
      h2 <- as.numeric(h2s[i, ])

      cat("h1 = (", paste(h1, collapse=", "), ")\n")
      cat("h2 = (", paste(h2, collapse=", "), ")\n")
      cat("P(h1)P(h2) =", prob.prod[i], "\n")

      if (place == contributors.place) {
        contributors.found <- TRUE
        cat("!!!!!!!!!! CONTRIBUTORS !!!!!!!!!!\n")
      }
      
      cat("--------------------------\n")
    }

    if (contributors.place > top.limit) {
      cat("The contributors were not in top-", top.limit, ".\n", sep="")
    } else {
      cat("The contributors were on ", contributors.place, ". place.\n", sep="")
    }

    cat("Sum of probabilities for possible contributors =", prob.prod.sum, "\n")
  }

  return(list(
    contributors.probs = contributors.probs,
    haplotype.universe.size = nrow(haplotype.universe),
    prob.sum = prob.prod.sum,
    contributors.place = contributors.place 
    ))
}
