library(MASS)
library(rpart)

#install.packages("kernlab")
library(kernlab)

#install.packages("e1071")
library(e1071)

#install.packages("Design")
library(Design)

##################################
##################################
####   Classifcation models   ####
##################################
##################################
# Given a formula, return a model

# Classification trees
rpart.classification.model.function <- function(formula, dataset) {
  return(rpart(formula, data=dataset))
}

rpart.tuning.classification.model.function <- function(formula, dataset) {
  tobj <- tune.rpart(formula, dataset, cp = c(0.0001, 0.001, 0.01))
  fit <- tobj$best.model  
  fit$performance <- tobj$best.performance  
  return(fit)
}

# Support vector machines - kernlab
svm.kernlab.classification.model.function <- function(formula, dataset) {
  return(ksvm(formula, 
              data = dataset, 
              type = "C-bsvc", 
              kernel = "rbfdot", 
              kpar = list(sigma = 0.1), 
              C = 10, 
              prob.model = TRUE))
}

# Support vector machines - e1071
svm.e1071.classification.model.function <- function(formula, dataset) {
  return(svm(formula, data = dataset, 
             method = "C-classification", 
             kernel = "radial", 
             cost = 10, 
             gamma = 0.1, 
             probability = TRUE))
}

svm.e1071.tuning.classification.model.function <- function(formula, dataset) {
  tobj <- tune.svm(formula, data = dataset, 
    gamma = 10^(-6:-3), 
    cost = 10^(1:2), 
    kernel = "radial", 
    probability = TRUE)

  fit <- tobj$best.model  

  # Carry performance along:
  fit$performance <- tobj$best.performance  
  
  return(fit)
}

# Ordered logistic regression
polr.classification.model.function <- function(formula, dataset) {
  return(polr(formula, data = dataset, method = "logistic", Hess = TRUE))
#  return(polr(formula, data = dataset, method = "cauchit", Hess = TRUE))
}

# Ordered logistic regression
lrm.classification.model.function <- function(formula, dataset) {
  return(lrm(formula, data = dataset))
}

##################################
##################################
####     Score functions      ####
##################################
##################################
# Given a list of models,
# return a vector of scores.
# The smaller score, the better.

generic.classification.model.score.iterator <- function(fit.list, score.function) {
  list.length <- length(fit.list)
  scores <- numeric(list.length)
  
  for (i in 1:list.length) {
    fit <- fit.list[[i]]
    scores[i] <- score.function(fit)
  }
  
  return(scores)
}

rpart.classification.model.score <- function(fit.list) {
  return(generic.classification.model.score.iterator(fit.list, 
    function(fit) {
      # Also checks for NULL
      if (!is(fit, "rpart")) stop("fit is not a rpart object")

      leaves <- which(fit$frame$var == "<leaf>")
      devar <- fit$frame$dev[leaves]
      return(sum(devar))
    }))
}

rpart.tuning.classification.model.score <- function(fit.list) {
  return(generic.classification.model.score.iterator(fit.list, 
    function(fit) {
      if (is.null(fit$performance)) stop("is.null(fit$performance)")
      return(fit$performance)
    }))
}

svm.kernlab.classification.model.score <- function(fit.list) {
  return(generic.classification.model.score.iterator(fit.list, 
    function(fit) {
      return(attributes(fit)$error)
    }))
}

svm.e1071.classification.model.score <- function(fit.list) {
  # This is a "hack": this is not that nice, but because the
  # dataset.name is changed globally, we can take advantage
  # to make a better generic framework (only this would need 
  # access to the dataset:
  dataset.extended <- get.extended.dataset(dataset.name)
  
  return(generic.classification.model.score.iterator(fit.list, 
    function(fit) {
      response.variable <- which(loci.names == as.character(fit$terms)[2])
      prediction <- predict(fit, dataset.extended)
      # FIXME: Must be done smarter
      prediction <- as.numeric(as.vector(prediction))
      result <- prediction - dataset.extended[, response.variable]
      return(as.numeric(crossprod(result)))
    }))
}

svm.e1071.tuning.classification.model.score <- function(fit.list) {
  return(generic.classification.model.score.iterator(fit.list, 
    function(fit) {
      if (is.null(fit$performance)) stop("is.null(fit$performance)")
      return(fit$performance)
    }))
}

polr.classification.model.score <- function(fit.list) {
  return(generic.classification.model.score.iterator(fit.list, 
    function(fit) {
      return(extractAIC(fit)[2])
    }))
}

lrm.classification.model.score <- function(fit.list) {
  return(generic.classification.model.score.iterator(fit.list, 
    function(fit) {
      # Minus because small scores are better
      return(-as.numeric(fit$stats[which(names(fit$stats) == "R2")]))
    }))
}

##################################
##################################
####   Prediction functions   ####
##################################
##################################
# Given a fit and a dataset, it
# predicts the probabilities

rpart.classification.model.prediction <- function(fit, dataset) {
  if (!is(fit, "rpart")) stop("fit is not a rpart object")  
  return(predict(fit, dataset))
}

svm.kernlab.classification.model.prediction <- function(fit, dataset) {
  return(predict(fit, dataset, type="probabilities"))
}

svm.e1071.classification.model.prediction <- function(fit, dataset) {
  prediction <- predict(fit, newdata=dataset, probability=TRUE)
  prediction <- attr(prediction, "probabilities")  
  return(prediction)
}

polr.classification.model.prediction <- function(fit, dataset) {
  return(predict(fit, dataset, type="p"))
}

lrm.classification.model.prediction <- function(fit, dataset) {
  mat <- predict(fit, dataset, type="fitted.ind")  
  colnames(mat) <- names(fit$freq)
  return(mat)
}

##################################
##################################
####     Definitions          ####
##################################
##################################
#classification.model.names <- c("rpart", "svm", "polr")
#classification.model.functions <- c(rpart.classification.model.function, 
#  svm.classification.model.function, polr.classification.model.function)
#classification.model.scores <- c(rpart.classification.model.score, 
#  svm.classification.model.scores, polr.classification.model.score)

# It's enough to just comment out the model name to exclude the model
classification.model.use <- c(1, 3, 4, 5, 6, 7, 8)
#classification.model.use <- c(1)

#classification.model.names <- c(
#  "rpart-no-tuning",                    # 1
#  "rpart-tuning-performance-score",     # 2
#  "rpart-tuning-deviance-score",        # 3
#  "svm-e1071-no-tuning",                # 4
#  "svm-e1071-tuning-performance-score", # 5
#  "svm-e1071-tuning-crossprod-score",   # 6
#  "polr",                               # 7
#  "lrm"                                 # 8
#)

classification.model.names <- c(
  "rpart",          # 1
  "rpart-tun-perf", # 2
  "rpart-tun-dev",  # 3
  "svm",            # 4
  "svm-tun-perf",   # 5
  "svm-tun-n",     # 6
  "polr",           # 7
  "lrm"             # 8
)

classification.model.functions <- c(
  rpart.classification.model.function, 
  rpart.tuning.classification.model.function, 
  rpart.tuning.classification.model.function, 
  svm.e1071.classification.model.function, 
  svm.e1071.tuning.classification.model.function, 
  svm.e1071.tuning.classification.model.function, 
  polr.classification.model.function,
  lrm.classification.model.function
)

classification.model.scores <- c(
  rpart.classification.model.score, 
  rpart.tuning.classification.model.score, 
  rpart.classification.model.score, 
  svm.e1071.classification.model.score,
  svm.e1071.tuning.classification.model.score,
  svm.e1071.classification.model.score,
  polr.classification.model.score,
  lrm.classification.model.score
)

classification.model.predictions <- c(
  rpart.classification.model.prediction, 
  rpart.classification.model.prediction,
  rpart.classification.model.prediction,
  svm.e1071.classification.model.prediction,
  svm.e1071.classification.model.prediction,
  svm.e1071.classification.model.prediction,
  polr.classification.model.prediction,
  lrm.classification.model.prediction
)

