surveying.regression <- function(W.group, dependent.group, start.values, include.beta1 = TRUE) {
  # nls: nonlinear least-squares

  #x11()
  #plot(W.group, dependent.group)

  control.list <- nls.control(maxiter=100, tol=1e-06, printEval=TRUE)
    
  if (include.beta1) {
    regression.model <- nls(dependent.group ~ beta1 + exp(beta2 * W.group + beta3), 
      start=start.values, trace=TRUE, control=control.list)
      
    print(summary(regression.model))
    return(summary(regression.model)$coefficients[, 1])
  } else {
    regression.model <- nls(dependent.group ~ exp(beta2 * W.group + beta3), 
      start=start.values, trace=TRUE, control=control.list)
      
    print(summary(regression.model))
    return(c(0, summary(regression.model)$coefficients[, 1]))
  }
}

surveying.analysis <- function(name, haplotypes.count, observation.count, Nis, W, W.group, mu.group, sigmasq.group, include.beta1 = TRUE) {
  if (include.beta1) {
    cat("Doing analysis with beta1, i.e. with formula: ~ beta1 + exp(beta2 * W.group + beta3).\n")
  } else {
    cat("Doing analysis without beta1, i.e. with formula: ~ exp(beta2 * W.group + beta3).\n")
  }
  
  descr <- paste(ifelse(include.beta1, "with", "without"), "-beta1", sep="")

  # Berlin
  #mu.start.values <- list(beta1 = 0.005, beta2 = 15, beta3 = -10)

  # Dane
  #mu.start.values <- list(beta1 = 0.0005, beta2 = 100, beta3 = -1)

  mu.start.values <- list(beta1 = 0.0005, beta2 = 100, beta3 = -1)
  
  sigmasq.start.values <- list(beta1 = -0.001, beta2 = 20, beta3 = -10)

  if (!include.beta1) {
    mu.start.values$beta1 <- NULL
    sigmasq.start.values$beta1 <- NULL
  }

  cat("mu-regression\n")
  beta.mu <- surveying.regression(W.group, mu.group, 
    start.values=mu.start.values, 
    include.beta1=include.beta1)

  cat("sigmasq-regression\n")
  beta.sigmasq <- surveying.regression(W.group, sigmasq.group, 
    start.values=sigmasq.start.values, 
    include.beta1=include.beta1)    

  if (beta.mu[1] < 0) {
    cat("\n!!! beta1 < 0 !!!\n")
    cat("Skipping...\n\n")
    return(FALSE)
  } 

  # beta for sigma^2 is called beta4 in the project
  if (beta.sigmasq[1] < 0) {
    cat("\n!!! beta4 < 0 !!!\n")
    cat("Skipping...\n\n")
    return(FALSE)
  } 
  
  # Calculating frequencies under the model
  mu <- beta.mu[1] + exp(beta.mu[2] * W + beta.mu[3])
  sigmasq <- beta.sigmasq[1] + exp(beta.sigmasq[2] * W + beta.sigmasq[3])

  if (any(mu < 0)) {
    cat("\n!!! Some frequencies had estimated mu less than zero !!!\n")
    cat("Estimated mu's:\n")
    print(mu[which(mu < 0)])
    cat("Indicies:\n")
    print(which(mu < 0))
    cat("Skipping...\n\n")
    return(FALSE)
  } 

  if (any(sigmasq < 0)) {
    cat("\n!!! Some frequencies had estimated sigmasq less than zero !!!\n")
    cat("Estimated sigmasq's:\n")
    print(sigmasq[which(sigmasq < 0)])
    cat("Indicies:\n")
    print(which(sigmasq < 0))
    cat("Skipping...\n\n")
    return(FALSE)
  }

  # Plotting
  mu.equation <- paste(round(beta.mu[1], 3), " + exp(", round(beta.mu[2], 3), " * W ", 
    ifelse(beta.mu[3] >= 0, paste("+", round(beta.mu[3], 3)), paste("-", round(-1*beta.mu[3], 3))), ")", sep="")
 
  pdf(paste("figures/surveying-", descr, "-", name, "-mu.pdf", sep=""))
  plot(W.group, mu.group, main=paste(name, "divided into", length(W.group), "groups"), xlab="W", ylab=expression(mu),
    sub = substitute(mu * (plain(W)) == muequationtxt, list(muequationtxt = mu.equation)))
  curve(beta.mu[1] + exp(beta.mu[2] * x + beta.mu[3]), add = TRUE)
  dev.off()

  sigmasq.equation <- paste(round(beta.sigmasq[1], 3), " + exp(", round(beta.sigmasq[2], 3), " * W ", 
    ifelse(beta.sigmasq[3] >= 0, paste("+", round(beta.sigmasq[3], 3)), paste("-", round(-1*beta.sigmasq[3], 3))), ")", sep="")

  pdf(paste("figures/surveying-", descr, "-", name, "-sigmasq.pdf", sep=""))
  plot(W.group, sigmasq.group, main=paste(name, "divided into", length(W.group), "groups"), xlab="W", ylab=expression(sigma^2), 
    sub = substitute(sigma^{2} * (plain(W)) == sigmasqequationtxt, list(parW = "(W)", sigmasqequationtxt = sigmasq.equation)))
  curve(beta.sigmasq[1] + exp(beta.sigmasq[2] * x + beta.sigmasq[3]), add = TRUE)
  dev.off()

  # Finds u_i and v_i for i = 1, ..., haplotypes.count
  u <- (mu^2 * (1 - mu)) / sigmasq
  v <- u * ( (1 - mu) / mu )

  # Expected singletons, doubletons etc.
  j.max <- 9
  js <- 0:j.max
  EK <- numeric(j.max+1)
  # FIXME
  #epsilon <- 10e-6
  #epsilon <- 10e-2

  for (j in js) {
    tmp <- 0
    for (i in 1:haplotypes.count) {
      tmp <- tmp + integrate(function(f) { 
          dbinom(j, observation.count-1, f) * dbeta(f, u[i], v[i])
#        }, lower=epsilon, upper=1-epsilon)$value
        }, lower=0, upper=1, stop.on.error=FALSE)$value
    }
    
    EK[j+1] <- tmp
  }

  trueK <- numeric(j.max+1)

  for (j in js) {
    trueK[j+1] <- length(which(Nis == j+1))
  }

  counts <- t(cbind(true=trueK, expected=EK))

  pdf(paste("figures/surveying-", descr, "-", name, "-observed-vs-expected-count.pdf", sep=""))
  bplot <- barplot(counts, main=paste(name, "based on", length(W.group), "groups"),
    xlab="j",
    ylab="Number of haplotypes observed j+1 times",  
    sub=paste("Using regression", ifelse(include.beta1, "with", "without"), "beta1 and beta4"),
    names=js,
    col=c("white","darkgrey"), 
    legend = c("Observed count", "Expected count"), 
    beside=TRUE)

  js.length <- length(js)
  labs <- numeric(js.length*2)
  labs[2*(1:js.length)-1] <- trueK
  labs[2*(1:js.length)] <- round(EK, 2)
  #text(bplot, par("usr")[3] + 2, srt = 270, adj = 1, labels = labs, xpd = TRUE, col="black")
  text(bplot, counts + 1, srt = 270, adj = 1, labels = labs, xpd = TRUE, col="black")
  dev.off()


  pdf(paste("figures/surveying-", descr, "-", name, "-observed-vs-expected-count-color.pdf", sep=""))
  bplot <- barplot(counts, main=paste(name, "based on", length(W.group), "groups"),
    xlab="j",
    ylab="Number of haplotypes observed j+1 times",  
    sub=paste("Using regression", ifelse(include.beta1, "with", "without"), "beta1 and beta4"),
    names=js,
    col=c(rgb(0.62,0.70,0.76), rgb(0,0.23,0.38)), 
    legend = c("Observed count", "Expected count"), 
    beside=TRUE)

  js.length <- length(js)
  labs <- numeric(js.length*2)
  labs[2*(1:js.length)-1] <- trueK
  labs[2*(1:js.length)] <- round(EK, 2)
  text(bplot, counts + 1, srt = 270, adj = 1, labels = labs, xpd = TRUE, col="black")
  dev.off()
  
  # Expected frequencies
  alpha <- u + Nis - 1
  beta <- v + observation.count - Nis
  posterior.means <- alpha / (alpha + beta)
  posterior.means.sum <- sum(posterior.means)
  probs <- posterior.means / posterior.means.sum
  expected.counts <- probs * observation.count

  dataset.probabilities <- data.frame(probsRoewer = posterior.means)   
  dataset.probabilities <- cbind(
    dataset.probabilities,
    probsRoewerNormalized = probs, 
    expectedRoewer = expected.counts)
  
#  dataset.probabilities <- cbind(
#    probsRoewer = posterior.means, 
#    probsRoewerNormalized = probs, 
#    expectedRoewer = expected.counts)

  pdf(paste("figures/surveying-", descr, "-", name, "-u-vs-v.pdf", sep=""))
  plot(u, v, 
    main=paste("fitted u vs. fitted v for", name, "based on", length(W.group), "groups"),
    sub=paste("Using regression", ifelse(include.beta1, "with", "without"), "beta1 and beta4"))
  dev.off()

  cat("Covered probability mass =", round(sum(dataset.probabilities$probsRoewer), 3), "\n")

  return(dataset.probabilities)
}
