library(dr)
library(xtable)

source("start.R")
source("include-surveying.R")
source("include-marginal-logic.R")

sink("log/surveying.log")

model.names <- c("Full regression model", "Reduced regression model")

unobserved.prob.matrix <- matrix(NA, 
  nrow=length(dataset.names), 
  ncol=2+length(model.names),
  dimnames=list(paste("¤DATASET:", dataset.names, "¤", sep=""), c("Estimate", "CI", model.names)))
unobserved.prob.estimates <- c("0.364", "0.602", "0.277")
unobserved.prob.intervals <- c("[0.321; 0.408]", "[0.510; 0.694]", "[0.212; 0.342]")

deviance.relative.predicted.matrix <- matrix(NA, 
  nrow=length(dataset.names), 
  ncol=length(model.names), 
  dimnames=list(paste("¤DATASET:", dataset.names, "¤", sep=""), model.names))

unobserved.include.beta1 <- numeric(length(dataset.names))
unobserved.exclude.beta1 <- numeric(length(dataset.names))

# Normally i, but here it is used so many other places
for (k in 1:length(dataset.names)) {
#for (i in 1:2) {
  dataset.name <- dataset.names[k]

  unobserved.prob.matrix[k, 1] <- unobserved.prob.estimates[k]
  unobserved.prob.matrix[k, 2] <- unobserved.prob.intervals[k]
  unobserved.prob.matrix[k, 3] <- NA # with beta1 and beta4
  unobserved.prob.matrix[k, 4] <- NA # without beta1 and beta4

  dataset.loci.count <- get.loci.count(dataset.name)
  
  cat("Starting surveying for dataset \"", dataset.name, "\".\n", sep="")

  #dataset <- get.extended.dataset(dataset.name)
  dataset <- get.compact.dataset(dataset.name)

  #haplotypes.count <- nrow(dataset)
  haplotypes.count <- get.haplotypes.count(dataset.name)
  
  #observation.count <- sum(dataset$N)
  observation.count <- get.observation.count(dataset.name)
  
  d <- as.matrix(dist(dataset[, 1:dataset.loci.count], method = "manhattan"))

  # Error check
  for (i in 1:haplotypes.count) {
    if (d[i, i] != 0) stop(paste("Vector must be at distance 0 from itself! Error at row", i))
    if (any(d[i,-i] == 0)) stop(paste("Vector must be at distance > 0 from other than itself! Error at row", i))
  }

  W <- numeric(haplotypes.count)

  # Highly readable, but not efficient!
  for (i in 1:haplotypes.count) {
    W[i] <- 0
    
    for (j in 1:haplotypes.count) {
      if (i == j) next      
      W[i] <- W[i] + dataset$N[j] / d[i, j]
    }
    W[i] <- W[i] / observation.count
  }
  
  slices <- 15
  
  W.slices <- dr.slices(W, slices)

  # nslices: Gives the actual number of slices produced, which may be
  #          smaller than the number requested.
  slices <- W.slices$nslices

  cat("Number of slices = ", slices, "\n")

  W.group <- numeric(slices)
  mu.group <- numeric(slices)
  sigmasq.group <- numeric(slices)

  for (i in 1:slices) {
    indices <- which(W.slices$slice.indicator == i)
    W.group[i] <- mean(W[indices])
    
    freqs <- (dataset$N[indices] - 1) / (observation.count - haplotypes.count)
    mu.group[i] <- mean(freqs)
    sigmasq.group[i] <- var(freqs)
  }

  dataset.probabilities = surveying.analysis(dataset.name, 
    haplotypes.count, observation.count, dataset$N, 
    W, W.group, mu.group, sigmasq.group, include.beta1 = TRUE)

  # If length(dataset.probabilities) > 1 then it is because the method succeeded
  if (length(dataset.probabilities) == 1 && dataset.probabilities == FALSE) {
    unobserved.include.beta1[k] <- NA
  } else {
    unobserved.prob.matrix[k, 3] <- round(1 - sum(dataset.probabilities$probsRoewer), 3) # with beta1 and beta4
    unobserved.include.beta1[k] <- 1 - sum(dataset.probabilities$probsRoewer)

    deviance.relative.predicted.matrix[k, 1] <- deviance.relfreq.predicted(dataset$N, dataset.probabilities$probsRoewer)
  }
  
  dataset.probabilities = surveying.analysis(dataset.name, 
    haplotypes.count, observation.count, dataset$N, 
    W, W.group, mu.group, sigmasq.group, include.beta1 = FALSE)

  # If length(dataset.probabilities) > 1 then it is because the method succeeded
  if (length(dataset.probabilities) == 1 && dataset.probabilities == FALSE) {
    unobserved.exclude.beta1[k] <- NA
  } else {
    unobserved.prob.matrix[k, 4] <- round(1 - sum(dataset.probabilities$probsRoewer), 3) # without beta1 and beta4
    unobserved.exclude.beta1[k] <- 1 - sum(dataset.probabilities$probsRoewer)

    deviance.relative.predicted.matrix[k, 2] <- deviance.relfreq.predicted(dataset$N, dataset.probabilities$probsRoewer)
  }
  
  cat("Dataset \"", dataset.name, "\" done.\n\n", sep="")
}

sink()

sink("out/unobserved-surveying.log")
for (k in 1:length(dataset.names)) {
  cat("Unobserved probability mass for dataset \"", dataset.names[k], 
    "\" including beta1 and beta4 = ", 
    round(unobserved.include.beta1[k], 3), "\n", sep="")  

  cat("Unobserved probability mass for dataset \"", dataset.names[k], 
    "\" including beta1 and beta4 = ", 
    round(unobserved.exclude.beta1[k], 3), "\n", sep="")  

  cat("\n")
}
sink()


unobserved.prob.matrix <- t(unobserved.prob.matrix)
tableOutput <- xtable(unobserved.prob.matrix,
      align=c("l", rep("r", ncol(unobserved.prob.matrix))),
      label=paste("tab:surveying-unobserved-probabilities", sep=""), 
      caption=paste("Unobserved probabilities using frequency surveying. The missing values indicate that the method did not succeed for that particular model."))
filename <- paste("out/surveying-unobserved-probabilities.tex", sep="")
print(tableOutput, file=filename)
do.replacements(filename)

deviance.relative.predicted.matrix <- t(deviance.relative.predicted.matrix)
tableOutput <- xtable(deviance.relative.predicted.matrix,
      align=c("l", rep("r", ncol(deviance.relative.predicted.matrix))),
      label=paste("tab:surveying-deviance-relative-predicted", sep=""), 
      caption=paste("Comparing deviance between relative frequencies and predicted probabilities found using frequency surveying. The missing values indicate that the method did not succeed for that particular model."))
filename <- paste("out/surveying-deviance-relative-predicted.tex", sep="")
print(tableOutput, file=filename)
do.replacements(filename)

